import 'dart:typed_data';

import 'package:bird/bird.dart';
import 'package:bird_cryptography/bird_cryptography.dart';
import 'package:bird_flutter/bird_flutter.dart';

// ignore_for_file: close_sinks
class BaseDigestBloc extends HookBloc {
  final Signal<CryptographicDataEncoding> sourceEncoding
  = HookBloc.disposeSink(Signal(e_hex));
  final Signal<CryptographicDataEncoding> targetEncoding
  = HookBloc.disposeSink(Signal(e_hex));
  final Signal<String> sourceText = HookBloc.disposeSink(Signal(""));
  final Signal<bool> isLoading = HookBloc.disposeSink(Signal(false));
  final Signal<String> targetText = HookBloc.disposeSink(Signal(""));
}

class DigestBloc extends BaseDigestBloc {
  Signal<SimpleDigest> _digest;

  DigestBloc(SimpleDigest digest) {
    _digest = Signal(digest);
    disposeSinkLater(_digest);
  }

  void convert() {
    isLoading.add(true);
    try {
      final _input = sourceEncoding.value.decode(sourceText.value);
      try {
        final _converted = _digest.value.digestRaw(Uint8List.fromList(_input));
        try {
          final _output = targetEncoding.value.encode(_converted);
          targetText.add(_output);
        } catch (e) {
          targetText.add("Invalid output for the encoding $e");
        }
      } catch (e) {
        targetText.add("Could not convert the input. $e");
      }
    } catch (e) {
      targetText.add("Invalid input for the encoding $e");
    }
    isLoading.add(false);
  }
}

class CryptoCipherPWIVBloc extends BaseDigestBloc {
  Signal<CryptoCipher<CryptoParametersPWAndIV>> _digest;
  final Signal<String> ivText = HookBloc.disposeSink(Signal(""));
  final Signal<CryptographicDataEncoding> ivEncoding
  = HookBloc.disposeSink(Signal(e_hex));
  final Signal<String> pwText = HookBloc.disposeSink(Signal(""));
  final Signal<CryptographicDataEncoding> pwEncoding
  = HookBloc.disposeSink(Signal(e_hex));

  CryptoCipherPWIVBloc(CryptoCipher<CryptoParametersPWAndIV> digest) {
    _digest = Signal(digest);
    disposeSinkLater(_digest);
  }

  void convert(bool encrypt) {
    isLoading.add(true);
    try {
      final _input = sourceEncoding.value.decode(sourceText.value);
      try {
        final _pw = pwEncoding.value.decode(pwText.value);
        try {
          final _iv = ivEncoding.value.decode(ivText.value);
          try {
            List<int> _converted;
            if (encrypt) {
              _converted = _digest.value.encrypt(
                _input,
                CryptoParametersPWAndIV.create(
                    Uint8List.fromList(_pw), Uint8List.fromList(_iv)),
              );
            } else {
              _converted = _digest.value.decrypt(
                _input,
                CryptoParametersPWAndIV.create(
                    Uint8List.fromList(_pw), Uint8List.fromList(_iv)),
              );
            }
            try {
              final _output = targetEncoding.value.encode(_converted);
              targetText.add(_output);
            } catch (e) {
              targetText.add("Invalid output/encoding $e");
            }
          } catch (e) {
            targetText.add("Could not convert the input. $e");
          }
        } catch (e) {
          targetText.add("Invalid iv/encoding $e");
        }
      } catch (e) {
        targetText.add("Invalid pw/encoding $e");
      }
    } catch (e) {
      targetText.add("Invalid input/encoding $e");
    }
    isLoading.add(false);
  }
}

class CryptoCipherPWBloc extends BaseDigestBloc {
  Signal<CryptoCipher<CryptoParametersPW>> _digest;
  final Signal<String> pwText = HookBloc.disposeSink(Signal(""));
  final Signal<CryptographicDataEncoding> pwEncoding
  = HookBloc.disposeSink(Signal(e_hex));

  CryptoCipherPWBloc(CryptoCipher<CryptoParametersPW> digest) {
    _digest = Signal(digest);
    disposeSinkLater(_digest);
  }

  void convert(bool encrypt) {
    isLoading.add(true);
    try {
      final _input = sourceEncoding.value.decode(sourceText.value);
      try {
        final _pw = pwEncoding.value.decode(pwText.value);
        try {
          List<int> _converted;
          if (encrypt) {
            _converted = _digest.value.encrypt(
              _input,
              CryptoParametersPW.create(Uint8List.fromList(_pw)),
            );
          } else {
            _converted = _digest.value.decrypt(
              _input,
              CryptoParametersPW.create(Uint8List.fromList(_pw)),
            );
          }
          try {
            final _output = targetEncoding.value.encode(_converted);
            targetText.add(_output);
          } catch (e) {
            targetText.add("Invalid output/encoding $e");
          }
        } catch (e) {
          targetText.add("Could not convert the input. $e");
        }
      } catch (e) {
        targetText.add("Invalid pw/encoding $e");
      }
    } catch (e) {
      targetText.add("Invalid input/encoding $e");
    }
    isLoading.add(false);
  }
}

class CryptoSimplePWDigestBloc extends BaseDigestBloc {
  Signal<SimplePWDigest> _digest;
  final Signal<String> pwText = HookBloc.disposeSink(Signal(""));
  final Signal<CryptographicDataEncoding> pwEncoding
  = HookBloc.disposeSink(Signal(e_hex));

  CryptoSimplePWDigestBloc(SimplePWDigest digest) {
    _digest = Signal(digest);
    disposeSinkLater(_digest);
  }

  void convert() {
    isLoading.add(true);
    try {
      final _input = sourceEncoding.value.decode(sourceText.value);
      try {
        final _pw = pwEncoding.value.decode(pwText.value);
        try {
          final _converted = _digest.value.digestRaw(
              Uint8List.fromList(_input), _pw);
          try {
            final _output = targetEncoding.value.encode(_converted);
            targetText.add(_output);
          } catch (e) {
            targetText.add("Invalid output/encoding $e");
          }
        } catch (e) {
          targetText.add("Could not convert the input. $e");
        }
      } catch (e) {
        targetText.add("Invalid pw/encoding $e");
      }
    } catch (e) {
      targetText.add("Invalid input/encoding $e");
    }
    isLoading.add(false);
  }
}
