import 'dart:typed_data';

import 'package:convert/convert.dart';
import 'package:meta/meta.dart';

/// https://en.wikipedia.org/wiki/Block_cipher_mode_of_operation#Cipher_Block_Chaining_(CBC)
/// XOR the IV with the plaintext and use the ciphertext as the IV for the next block.
/// XOR the IV with the ciphertext and use the paintext as the IV for the next block.
class BlockModeCBC {
  final int blockSize;

  final Uint8List iv;

  final Uint8List Function(Uint8List) cipher;

  const BlockModeCBC(this.blockSize, this.iv, this.cipher);

  Uint8List crypt(Uint8List data, bool encrypt) {
    Uint8List lastBlock;

    final cryptedBlocks = divideIntoBlocks(data, blockSize)
        .fold<List<Uint8List>>([],
            (List<Uint8List> cum, Uint8List currentBlock) {
      final Uint8List curIV =
          cum.isEmpty ? iv : (encrypt == true ? cum.last : lastBlock);
      if (encrypt) {
        cum.add(cipher(
            xorbytes(currentBlock, curIV, blockSize, PaddingLocation.right)));
      } else {
        cum.add(xorbytes(
            cipher(currentBlock), curIV, blockSize, PaddingLocation.right));
      }
      lastBlock = currentBlock;
      return cum;
    });

    return Uint8List.fromList(
        cryptedBlocks.fold<List<int>>([], (a, b) => a..addAll(b)));
  }
}

/// https://en.wikipedia.org/wiki/Block_cipher_mode_of_operation#Counter_(CTR)
/// IV is incremented on every block. starts out with an increment of 0
/// first block = 0
/// second block = 1
/// ...
class BlockModeCTR {
  final int blockSize;

  final Uint8List iv;

  final Uint8List Function(Uint8List) encrypt;

  const BlockModeCTR({
    @required this.blockSize,
    @required this.encrypt,
    @required this.iv,
  });

  Uint8List crypt(Uint8List data) {
    var blockCount = BigInt.zero;
    final initialIV = BigInt.parse(hex.encode(iv), radix: 16);
    final cryptedBlocks = divideIntoBlocks(data, blockSize)
        .fold<List<Uint8List>>([],
            (List<Uint8List> cum, Uint8List currentBlock) {
      final BigInt newBlockCount = initialIV + blockCount;

      /// pad because a value like "aef..." will not be "0aef"
      final currentIV =
          hex.decode(newBlockCount.toRadixString(16).padLeft(32, "0"));

      final ivThroughCipher = encrypt(Uint8List.fromList(currentIV));

      blockCount += BigInt.one;

      return cum
        ..add(xorbytes(
            ivThroughCipher, currentBlock, blockSize, PaddingLocation.right));
    });
    return Uint8List.fromList(
        cryptedBlocks.fold<List<int>>([], (a, b) => a..addAll(b)));
  }
}

class BlockModeECB {
  final int blockSize;

  final Uint8List Function(Uint8List) cipher;

  const BlockModeECB({
    @required this.blockSize,
    @required this.cipher,
  });

  Uint8List crypt(Uint8List data) {
    final cryptedBlocks = divideIntoBlocks(data, blockSize)
        .fold<List<Uint8List>>([],
            (List<Uint8List> cum, Uint8List currentBlock) {
      final padded = padbytes(currentBlock, PaddingLocation.right, blockSize);
      return cum..add(cipher(padded));
    });
    return Uint8List.fromList(
        cryptedBlocks.fold<List<int>>([], (a, b) => a..addAll(b)));
  }
}

/// Splits data into blocks of size [blockSize].
///
/// The last block can be smaller than [blockSize]
/// when data is not divisible into equal blocks of size blockSize.
List<Uint8List> divideIntoBlocks(Uint8List data, int blockSize) {
  var origData = Uint8List.fromList(data);
  final List<Uint8List> blocks = [];

  while (origData.isNotEmpty) {
    if (origData.length >= blockSize) {
      blocks.add(Uint8List.fromList(origData.sublist(0, blockSize)));
      origData =
          Uint8List.fromList(origData.sublist(blockSize, origData.length));
    } else {
      blocks.add(origData);
      origData = Uint8List.fromList([]);
    }
  }

  return blocks;
}

List<int> pad(List<int> list, PaddingLocation location, int length,
    {int symbol = 0}) {
  final List<int> result = List.from(list);

  if (result.length > length) {
    switch (location) {
      case PaddingLocation.left:
        return result.sublist(0, length);
        break;
      case PaddingLocation.right:
        return result.sublist(result.length - length, result.length);
        break;
    }
  } else {
    while (result.length < length) {
      switch (location) {
        case PaddingLocation.left:
          result.insert(0, symbol);
          break;
        case PaddingLocation.right:
          result.insert(result.length, symbol);
          break;
      }
    }
  }
  return result;
}

Uint8List padbytes(Uint8List list, PaddingLocation location, int length,
        {int symbol = 0}) =>
    Uint8List.fromList(pad(list, location, length, symbol: symbol));

List<int> xor(List<int> first, List<int> second, int length,
    PaddingLocation paddingLocation,
    {int symbol = 0}) {
  final firstList =
      pad(List.from(first), paddingLocation, length, symbol: null);
  final secondList =
      pad(List.from(second), paddingLocation, length, symbol: null);

  final List<int> result = List(length);

  for (int i = 0; i < length; i++) {
    if (firstList[i] == null && secondList[i] == null) {
      result[i] = symbol;
    } else if (firstList[i] == null) {
      result[i] = secondList[i];
    } else if (secondList[i] == null) {
      result[i] = firstList[i];
    } else {
      result[i] = firstList[i] ^ secondList[i];
    }
  }

  return result;
}

Uint8List xorbytes(Uint8List first, Uint8List second, int length,
        PaddingLocation paddingLocation, {int symbol = 0}) =>
    Uint8List.fromList(
        xor(first, second, length, paddingLocation, symbol: symbol));

enum PaddingLocation { left, right }
