import 'dart:convert';

import 'package:convert/convert.dart';
import 'package:bird_cryptography/bird_cryptography.dart';

const CryptographicDataEncoding e_base64 = CryptoEncodingBase64();
const CryptographicDataEncoding e_bytelist = CryptoEncodingByteList();
const CryptographicDataEncoding e_hex = CryptoEncodingHex();
const CryptographicDataEncoding e_unknown = CryptoEncodingUnknown();
const CryptographicDataEncoding e_intlist = CryptoEncodingIntList();
const CryptographicDataEncoding e_utf8 = CryptoEncodingUTF8();
const CryptographicDataEncoding e_ascii = CryptoEncodingASCII();

/// Describes common encodings used in operations of cryptographic primitives.
abstract class CryptographicDataEncoding extends Codec<List<int>, String> {
  const CryptographicDataEncoding();

  @override
  List<int> decode(String encoded) {
    if (encoded == null) {
      return null;
    }
    return super.decode(encoded);
  }

  @override
  String encode(List<int> decoded) {
    if (decoded == null) {
      return null;
    }
    return super.encode(decoded);
  }

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is CryptographicDataEncoding &&
              runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;

  FocusedCodec decodedData(List<int> raw) => FocusedCodec.raw(this, raw);

  FocusedCodec encodedData(String raw) => FocusedCodec.encoded(this, raw);
}

class CryptoEncodingBase64 extends CryptographicDataEncoding {
  const CryptoEncodingBase64();

  @override
  Converter<String, List<int>> get decoder => base64.decoder;

  @override
  Converter<List<int>, String> get encoder => base64.encoder;

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          other is CryptoEncodingBase64 &&
              runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

class CryptoEncodingByteList extends CryptographicDataEncoding {
  const CryptoEncodingByteList();

  static bool isListOfBytes(List<int> raw) =>
      !raw.map((i) => i > 255 || i.isNegative).contains(true);

  @override
  Converter<String, List<int>> get decoder =>
      _AnonymousConverter((raw) {
        assert(raw != null);
        if (raw == "[]") return [];
        final parsed = e_intlist.decode(raw);
        if (isListOfBytes(parsed)) {
          return parsed;
        } else {
          return null;
        }
      });

  @override
  Converter<List<int>, String> get encoder =>
      _AnonymousConverter((raw) {
        assert(raw != null);
        assert(isListOfBytes(raw));
        return "[${raw.join(", ")}]";
      });

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          super == other &&
              other is CryptoEncodingByteList &&
              runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

class CryptoEncodingHex extends CryptographicDataEncoding {
  const CryptoEncodingHex();

  @override
  Converter<String, List<int>> get decoder =>
      _AnonymousConverter((raw) {
        assert(raw != null);
        if (raw.startsWith("0x")) {
          return hex.decode(raw.substring(2));
        } else {
          return hex.decode(raw);
        }
      });

  @override
  Converter<List<int>, String> get encoder =>
      _AnonymousConverter((raw) {
        assert(raw != null);
        return hex.encode(raw);
      });

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          super == other &&
              other is CryptoEncodingHex &&
              runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

class CryptoEncodingIntList extends CryptographicDataEncoding {
  const CryptoEncodingIntList();

  @override
  Converter<String, List<int>> get decoder =>
      _AnonymousConverter((raw) {
        assert(raw != null);
        return List<dynamic>.from(json.decode(raw) as List).cast<int>();
      });

  @override
  Converter<List<int>, String> get encoder =>
      _AnonymousConverter((raw) {
        if (raw == null) return null;
        return "[${raw.join(", ")}]";
      });

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          super == other &&
              other is CryptoEncodingIntList &&
              runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

class CryptoEncodingUnknown extends CryptographicDataEncoding {
  const CryptoEncodingUnknown();

  @override
  Converter<String, List<int>> get decoder =>
      _AnonymousConverter((raw) => null);

  @override
  Converter<List<int>, String> get encoder =>
      _AnonymousConverter((raw) => null);

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          super == other &&
              other is CryptoEncodingUnknown &&
              runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

class CryptoEncodingUTF8 extends CryptographicDataEncoding {
  const CryptoEncodingUTF8();

  @override
  Converter<String, List<int>> get decoder => utf8.encoder;

  @override
  Converter<List<int>, String> get encoder => utf8.decoder;

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          super == other &&
              other is CryptoEncodingUTF8 &&
              runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

class CryptoEncodingASCII extends CryptographicDataEncoding {
  const CryptoEncodingASCII();

  @override
  Converter<String, List<int>> get decoder => ascii.encoder;

  @override
  Converter<List<int>, String> get encoder => ascii.decoder;

  @override
  bool operator ==(Object other) =>
      identical(this, other) ||
          super == other &&
              other is CryptoEncodingASCII &&
              runtimeType == other.runtimeType;

  @override
  int get hashCode => 0;
}

class _AnonymousConverter<S, T> extends Converter<S, T> {
  final T Function(S) _convert;

  const _AnonymousConverter(this._convert);

  @override
  T convert(S input) => _convert(input);
}
